"""`Adict` is an attribute-accessible dynamic dict wrapper, which allows to access dict items in
attribute notation (".") and allows friendly checks for non-existing items.
"""

from collections.abc import MutableMapping
import json
from typing import Any, Dict, Iterator, Optional, TYPE_CHECKING


__NONE__ = object()


# only py>=39 supports correct type hinting MutableMapping[Any, Any]
class Adict(MutableMapping):  # type: ignore
    """Attribute-accessible implementation of a dictionary."""

    def __init__(self, base: Optional[Dict[Any, Any]] = None, **kwargs: Any):
        """Construct a dynamic dictionary from an existing or an empty base dictionary.

        Args:
            base (dict): The dictionary from which the dynamic dict is constructed from.
        """
        self.__dict: Dict[Any, Any] = base if base is not None else {}
        self.__dict.update(kwargs)

    # dict function overrides

    def __getitem__(self, key: Any) -> Any:
        """Return items using the [] notation from the contained dict. """
        item = self.__getattr__(key)
        if item is None:
            raise KeyError(key)
        return item

    def __setitem__(self, key: Any, value: Any) -> None:
        """Write items using the [] notation into the contained dict."""
        self.__dict[key] = value

    def __delitem__(self, key: Any) -> None:
        del self.__dict[key]

    def __iter__(self) -> Iterator[Any]:
        return iter(self.__dict)

    def __len__(self) -> int:
        return len(self.__dict)

    # object overrides to make the attribute accessible

    def __getattr__(self, item: Any) -> Any:
        """Retrieve the value for the given attribute name from the dict.

        When the value is a dictionary, it is wrapped in another Adict instance.

        Note: __getattr__ is only for the attributes, not defined by the class,
        __getattribute__ catches all atrtibutes.
        """
        if item in self.__dict:
            value = self.__dict[item]
            if isinstance(value, dict):
                return Adict(value)
            return value
        return None

    def __setattr__(self, key: str, value: Any) -> None:
        """Store all attributes in the dictionary, except the explicit attributes.

        Note: Different from __getattr__(), __setattr__() overwrites _all_ attributes setting,
              not only the ones which don't have an object member. So __setattr__() is the
              counterpart of __getattribute__().
        """
        if key.startswith(f'_{type(self).__name__}__'):
            super().__setattr__(key, value)
        else:
            if isinstance(value, dict):
                self.__dict[key] = Adict(value)
            else:
                self.__dict[key] = value

    def __delattr__(self, key: str) -> None:
        if key in self.__dict:
            del self.__dict[key]

    # bool and str overrides

    def __bool__(self) -> bool:
        return bool(self.__dict)

    def __str__(self) -> str:
        """String is the dictionoary as string."""
        return self.__dict.__str__()

    def __repr__(self) -> str:
        """Representation is the dictionaries representation."""
        return self.__dict.__repr__()

    if TYPE_CHECKING:  # pragma: no cover
        __dict__ = {}  # type: Dict[Any, Any]
    else:
        @property
        def __dict__(self) -> Dict[Any, Any]:
            """Dict representation is the contained dictonary."""
            return self.__dict

    # safe accessing
    def __call__(self, key: str, default: Any = __NONE__) -> Any:
        value = self.__getattr__(key)
        if value is not None:
            return value

        if default is not __NONE__:
            if isinstance(default, dict):
                return Adict(default)
            return default
        return Adict()


class JsonEncoder(json.encoder.JSONEncoder):
    """Basic Json Encoder, which transforms any Adict elements into back into their __dict__
       representation. Additionally, we transform any other types into their str()
       representation."""

    def default(self, o: object) -> Any:
        if isinstance(o, Adict):
            return o.__dict__
        return str(o)



import setuptools

setuptools.setup(
    name='dictat',
    version='1.1.0',
    description='Adict is an attribute-accessible dynamic dictionary wrapper',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Janos Kutscherauer',
    author_email='janoskut@gmail.com',
    url='https://gitlab.com/janoskut/dictat',
    license_files = ['UNLICENSE'],
    packages=setuptools.find_packages(),
    install_requires=[],
    python_requires='>=3.7',
)



"""`pytest` unit tests for module `dictat` and it's classes."""

# pylint: disable=redefined-outer-name
# pylint: disable=missing-function-docstring

from typing import Any, Dict
import json

import pytest

from dictat import Adict, JsonEncoder


@pytest.fixture
def adict() -> Adict:
    return Adict()


@pytest.fixture
def subdict() -> Dict[Any, Any]:
    return {
        'key1': 'val1',
        'key2': 'val2',
        'subsub': {
            'subkey1': 'subval1',
            'subkey2': 'subval2',
        }
    }


def test_access_non_existing(adict: Adict) -> None:

    assert 'key' not in adict
    assert ('key' in adict) is False

    # default dict behaviour
    with pytest.raises(KeyError):
        adict['key']  # pylint: disable=pointless-statement

    # safe access with attribute notation
    assert adict.key is None


def test_bool(adict: Adict) -> None:

    assert not adict
    assert bool(adict) is False

    adict.key = 'value'
    assert adict
    assert bool(adict) is True

    adict = Adict()
    adict['key'] = 'value'
    assert adict
    assert bool(adict) is True


def test_len(adict: Adict) -> None:

    assert len(adict) == 0
    adict.key = 'value'
    assert len(adict) == 1
    adict.key2 = 'value2'
    assert len(adict) == 2
    adict.sub = {}
    assert isinstance(adict.sub, Adict)  # check needed to convince type checker
    adict.sub.key = 'subval'
    assert len(adict) == 3


def test_iterator(adict: Adict) -> None:
    count = 0
    for _ in adict:
        count += 1
    assert count == 0

    adict.a = 1
    adict.b = 2
    adict.c = 3

    keys = []
    sum_ = 0
    for key in adict:
        keys.append(key)
        sum_ += adict[key]
    assert keys == ['a', 'b', 'c']
    assert sum_ == 6


def test_get_set(adict: Adict) -> None:

    adict.key = 'value'
    assert adict.key == 'value'
    assert adict['key'] == 'value'

    adict.sub = {}
    # technical detail: dicts are returned wrapped in an 'Adict'
    assert isinstance(adict.sub, Adict)
    assert type(adict.sub) is Adict  # pylint: disable=unidiomatic-typecheck
    assert type(adict['sub']) is Adict  # pylint: disable=unidiomatic-typecheck

    assert adict.sub.key is None
    # default dict behaviour
    with pytest.raises(KeyError):
        adict.sub['key']  # pylint: disable=pointless-statement


def test_sub_dict(adict: Adict, subdict: Dict[Any, Any]) -> None:
    adict.sub = subdict
    assert isinstance(adict.sub, Adict)  # check needed to convince type checker
    assert adict.sub.key1 == 'val1'
    assert adict.sub.key2 == 'val2'
    assert adict.sub.subsub.subkey1 == 'subval1'
    assert adict.sub.subsub.subkey2 == 'subval2'


def test_paren_syntax(adict: Adict, subdict: Dict[Any, Any]) -> None:
    """Test safe access to non-existing keys using the paranthesis syntax.

    The only thing that can raise from here, is if we've traversed down to a leaf value, which
    then cannot accessed using () syntax."""
    adict.sub = subdict

    assert adict('sub')('key1') == 'val1'
    assert adict('sub')('key2') == 'val2'
    assert adict('sub')('subsub')('subkey1') == 'subval1'
    assert adict('sub')('subsub')('subkey2') == 'subval2'

    assert not adict('a')('b')('c')
    assert adict('a')('b')('c') == {}
    assert not adict('sub')('key3')
    assert adict('sub')('key3') == {}

    # with default value, e.g. when we know the last level is a leaf level
    assert adict('a')('b')('c', default=None) is None
    assert adict('sub')('key3', default=None) is None
    assert adict('a')('b')('c', 'd') == 'd'
    # test that the default=__NONE__=object() semantic is correct
    assert adict('a')('b')('c', default=object()) is not None
    assert adict('a')('b')('c', default=object()) != {}

    # equivalence of `dict.get('key', {})`, but converting `{}` to `Adict({})``
    assert adict('a', default={}) == {}
    assert isinstance(adict('a', default={}), Adict)
    assert adict('a', default={'b': 'c'})('b') == 'c'

    # the only chance to raise is when we ()-access a leaf value
    with pytest.raises(TypeError) as info:
        adict('sub')('subsub')('subkey2')('nomore')
    assert "'str' object is not callable" in str(info)


def test_delete(adict: Adict) -> None:

    # non-existing
    with pytest.raises(KeyError):
        del adict['key']

    # non-existing, attribute notation: safe to use
    del adict.key

    adict.key = 'value'
    assert adict.key == 'value'
    del adict.key
    assert 'key' not in adict
    assert adict.key is None


def test_mutable(adict: Adict) -> None:

    sub: Dict[Any, Any] = {}
    adict.sub = sub
    assert isinstance(adict.sub, Adict)  # check needed to convince type checker
    sub['key'] = 'value'
    assert adict.sub.key == 'value'

    adict.sub['new'] = 'newval'
    assert sub['new'] == 'newval'

    sub['key'] = 'changed'
    assert adict.sub.key == 'changed'


def test_initial_dict(subdict: Dict[Any, Any]) -> None:
    adict = Adict(subdict)
    assert adict.key1 == 'val1'
    assert adict.key2 == 'val2'
    assert adict.subsub.subkey1 == 'subval1'
    assert adict.subsub.subkey2 == 'subval2'

    assert adict.__dict__ == subdict
    assert adict.__dict__ is subdict


def test_initial_kwargs() -> None:
    adict = Adict(key3='val3', key4='val4')
    assert adict.key3 == 'val3'
    assert adict.key4 == 'val4'


def test_initial_dict_and_kwargs(subdict: Dict[Any, Any]) -> None:
    adict = Adict(subdict, key3='val3', key4='val4')
    assert adict.key1 == 'val1'
    assert adict.key2 == 'val2'
    assert adict.subsub.subkey1 == 'subval1'
    assert adict.subsub.subkey2 == 'subval2'
    assert adict.key3 == 'val3'
    assert adict.key4 == 'val4'


def test_str(adict: Adict, subdict: Dict[Any, Any]) -> None:

    assert str(adict) == '{}'

    adict.key = 'value'
    assert str(adict) == str({'key': 'value'})

    adict.sub = subdict
    assert str(adict) == str({'key': 'value', 'sub': subdict})


def test_dict(adict: Adict, subdict: Dict[Any, Any]) -> None:

    assert adict.__dict__ == {}

    adict.key = 'value'
    assert adict.__dict__ == {'key': 'value'}

    adict.sub = subdict
    assert adict.__dict__ == {'key': 'value', 'sub': subdict}


def test_json(adict: Adict, subdict: Dict[Any, Any]) -> None:

    import pathlib  # pylint: disable=import-outside-toplevel

    adict.sub = subdict
    adict.path = pathlib.Path('somepath')  # normally not JSON serializable

    # equivalent standad dict
    standard = {
        'sub': subdict,
        'path': str(pathlib.Path('somepath'))
    }

    dump_adict = json.dumps(adict, cls=JsonEncoder)
    dump_standard = json.dumps(standard)
    assert dump_adict == dump_standard

    # the encoder can also be used for standard dicts
    dump_standard2 = json.dumps(standard, cls=JsonEncoder)
    assert dump_standard2 == dump_standard
    assert dump_standard2 == dump_adict
